
import logging
import warnings
import sys
_log = logging.getLogger(__name__)

import asyncio

from functools import partial, wraps

from . import raw
from .raw import Disconnected, RemoteError, Cancelled, Finished, LazyRepr
from ..wrapper import Value, Type
from .._p4p import (logLevelAll, logLevelTrace, logLevelDebug,
                    logLevelInfo, logLevelWarn, logLevelError,
                    logLevelFatal, logLevelOff)

__all__ = [
    'Context',
    'Value',
    'Type',
    'RemoteError',
    'timeout',
]


def timesout(deftimeout=5.0):
    """Decorate a coroutine to implement an overall timeout.

    The decorated coroutine will have an additional keyword
    argument 'timeout=' which gives a timeout in seconds,
    or None to disable timeout.

    :param float deftimeout: The default timeout= for the decorated coroutine.

    It is suggested perform one overall timeout at a high level
    rather than multiple timeouts on low-level operations. ::

        @timesout()
        @asyncio.coroutine
        def dostuff(ctxt):
            yield from ctxt.put('msg', 'Working')
            A, B = yield from ctxt.get(['foo', 'bar'])
            yield from ctxt.put('bar', A+B, wait=True)
            yield from ctxt.put('msg', 'Done')

        @asyncio.coroutine
        def exec():
            with Context('pva') as ctxt:
                yield from dostuff(ctxt, timeout=5)
    """
    def decorate(fn):
        assert asyncio.iscoroutinefunction(fn), "Place @timesout before @coroutine"

        @wraps(fn)
        @asyncio.coroutine
        def wrapper(*args, timeout=deftimeout, **kws):
            loop = kws.get('loop')
            fut = fn(*args, **kws)
            if timeout is None:
                yield from fut
            else:
                yield from asyncio.wait_for(fut, timeout=timeout, loop=loop)
        return wrapper
    return decorate


class Context(raw.Context):

    """
    :param str provider: A Provider name.  Try "pva" or run :py:meth:`Context.providers` for a complete list.
    :param conf dict: Configuration to pass to provider.  Depends on provider selected.
    :param bool useenv: Allow the provider to use configuration from the process environment.
    :param dict nt: Controls :ref:`unwrap`.  None uses defaults.  Set False to disable
    :param dict unwrap: Legacy :ref:`unwrap`.

    The methods of this Context will block the calling thread until completion or timeout

    The meaning, and allowed keys, of the configuration dictionary depend on the provider.

    The "pva" provider understands the following keys:

    * EPICS_PVA_ADDR_LIST
    * EPICS_PVA_AUTO_ADDR_LIST
    * EPICS_PVA_SERVER_PORT
    * EPICS_PVA_BROADCAST_PORT

    Timeout and Cancellation
    ^^^^^^^^^^^^^^^^^^^^^^^^

    All coroutines/Futures returned by Context methods can be cancelled.
    The methods of Context do not directly implement a timeout.
    Instead :py:meth:`asyncio.wait_for` should be used.
    It is suggested perform one overall timeout at a high level
    rather than multiple timeouts on low-level operations. ::

        @timesout()
        @asyncio.coroutine
        def dostuff(ctxt):
            yield from ctxt.put('msg', 'Working')
            A, B = yield from ctxt.get(['foo', 'bar'])
            yield from ctxt.put('bar', A+B, wait=True)
            yield from ctxt.put('msg', 'Done')

        @asyncio.coroutine
        def exec():
            with Context('pva') as ctxt:
                yield from dostuff(ctxt, timeout=5)
    """

    def __init__(self, provider='pva', conf=None, useenv=True, nt=None, unwrap=None,
                 loop=None):
        super(Context, self).__init__(provider, conf=conf, useenv=useenv, nt=nt, unwrap=unwrap)
        self.loop = loop or asyncio.get_event_loop()

    @asyncio.coroutine
    def get(self, name, request=None):
        """Fetch current value of some number of PVs.

        :param name: A single name string or list of name strings
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.

        :returns: A p4p.Value, or list of same.  Subject to :py:ref:`unwrap`.

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values. ::

            with Context('pva') as ctxt:
                V    = yield from ctxt.get('pv:name')
                A, B = yield from ctxt.get(['pv:1', 'pv:2'])
        """
        singlepv = isinstance(name, (bytes, str))
        if singlepv:
            return (yield from self._get_one(name, request=request))

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)

        futs = [self._get_one(N, request=R) for N, R in zip(name, request)]

        ret = yield from asyncio.gather(*futs, loop=self.loop)

        return ret

    @asyncio.coroutine
    def _get_one(self, name, request=None):
        F = asyncio.Future(loop=self.loop)

        def cb(value):
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, (RemoteError, Disconnected, Cancelled)):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(self.loop.call_soon_threadsafe, cb)

        op = super(Context, self).get(name, cb, request=request)

        _log.debug('get %s request=%s', name, request)
        try:
            return (yield from F)
        finally:
            op.close()

    @asyncio.coroutine
    def put(self, name, values, request=None, process=None, wait=None, get=True):
        """Write a new value of some number of PVs.

        :param name: A single name string or list of name strings
        :param values: A single value, a list of values, a dict, a `Value`.  May be modified by the constructor nt= argument.
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param str process: Control remote processing.  May be 'true', 'false', 'passive', or None.
        :param bool wait: Wait for all server processing to complete.
        :param bool get: Whether to do a Get before the Put.  If True then the value passed to the builder callable
                         will be initialized with recent PV values.  eg. use this with NTEnum to find the enumeration list.

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values

        If 'wait' or 'process' is specified, then 'request' must be omitted or None. ::

            with Context('pva') as ctxt:
                yield from ctxt.put('pv:name', 5.0)
                yield from ctxt.put(['pv:1', 'pv:2'], [1.0, 2.0])
                yield from ctxt.put('pv:name', {'value':5})

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        if request and (process or wait is not None):
            raise ValueError("request= is mutually exclusive to process= or wait=")
        elif process or wait is not None:
            request = 'field()record[block=%s,process=%s]' % ('true' if wait else 'false', process or 'passive')

        singlepv = isinstance(name, (bytes, str))
        if singlepv:
            return (yield from self._put_one(name, values, request=request, get=get))

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)
        assert len(name) == len(values), (name, values)

        futs = [self._put_one(N, V, request=R, get=get) for N, V, R in zip(name, values, request)]

        yield from asyncio.gather(*futs, loop=self.loop)

    @asyncio.coroutine
    def _put_one(self, name, value, request=None, get=True):
        F = asyncio.Future(loop=self.loop)

        def cb(value):
            _log.debug("put done %s %s", name, LazyRepr(value))
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, (RemoteError, Disconnected, Cancelled)):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(self.loop.call_soon_threadsafe, cb)

        op = super(Context, self).put(name, cb, builder=value, request=request, get=get)

        _log.debug('put %s <- %s request=%s', name, LazyRepr(value), request)
        try:
            value = yield from F
        finally:
            op.close()

    @asyncio.coroutine
    def rpc(self, name, value, request=None):
        """Perform a Remote Procedure Call (RPC) operation

        :param str name: PV name string
        :param Value value: Arguments.  Must be Value instance
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.

        :returns: A Value.  Subject to :py:ref:`unwrap`.

        For example: ::

            uri = NTURI(['A','B'])
            with Context('pva') as ctxt:
                result = yield from ctxt.rpc('pv:name:add', uri.wrap('pv:name:add', 5, B=6))

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict or Value, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        F = asyncio.Future(loop=self.loop)

        def cb(value):
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, (RemoteError, Disconnected, Cancelled)):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(self.loop.call_soon_threadsafe, cb)

        op = super(Context, self).rpc(name, cb, value, request=request)

        try:
            return (yield from F)
        finally:
            op.close()

    def monitor(self, name, cb, request=None, notify_disconnect=False):
        """Create a subscription.

        :param str name: PV name string
        :param callable cb: Processing callback
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param bool notify_disconnect: In additional to Values, the callback may also be call with instances of Exception.
                                       Specifically: Disconnected , RemoteError, or Cancelled
        :returns: a :py:class:`Subscription` instance

        The callable will be invoked with one argument which is either.

        * A p4p.Value (Subject to :py:ref:`unwrap`)
        * A sub-class of Exception (Disconnected , RemoteError, or Cancelled)
        """
        assert asyncio.iscoroutinefunction(cb), "monitor callback must be coroutine"
        R = Subscription(name, cb, notify_disconnect=notify_disconnect, loop=self.loop)
        cb = partial(self.loop.call_soon_threadsafe, R._event)

        R._S = super(Context, self).monitor(name, cb, request)
        return R


class Subscription(object):

    """An active subscription.
    """

    def __init__(self, name, cb, notify_disconnect=False, loop=None):
        self.name, self._S, self._cb, self.loop = name, None, cb, loop
        self._notify_disconnect = notify_disconnect

        self._Q = asyncio.Queue(loop=self.loop)

        if notify_disconnect:
            self._Q.put_nowait(Disconnected())  # all subscriptions are inittially disconnected

        self._T = self.loop.create_task(self._handle())

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.close()

    def close(self):
        """Begin closing subscription.
        """
        if self._S is not None:
            # after .close() self._event should never be called
            self._S.close()
            self._S = None
            self._Q.put_nowait(None)

    @property
    def done(self):
        'Has all data for this subscription been received?'
        return self._S is None or self._S.done()

    @property
    def empty(self):
        'Is data pending in event queue?'
        return self._S is None or self._S.empty()

    @asyncio.coroutine
    def wait_closed(self):
        """Wait until subscription is closed.
        """
        assert self._S is None, "Not close()'d"
        yield from self._T

    def _event(self, value):
        if self._S is not None:
            self._Q.put_nowait(value)

    @asyncio.coroutine
    def _handle(self):
        E = None
        try:
            while True:
                E = yield from self._Q.get()
                self._Q.task_done()
                _log.debug("Subscription %s handle %s", self.name, LazyRepr(E))

                S = self._S

                if isinstance(E, Cancelled):
                    return

                elif isinstance(E, Disconnected):
                    _log.debug('Subscription notify for %s with %s', self.name, E)
                    if self._notify_disconnect:
                        yield from self._cb(E)
                    else:
                        _log.info("Subscription disconnect %s", self.name)
                    continue

                elif isinstance(E, RemoteError):
                    _log.debug('Subscription notify for %s with %s', self.name, E)
                    if self._notify_disconnect:
                        yield from self._cb(E)
                    elif isinstance(E, RemoteError):
                        _log.error("Subscription Error %s", E)
                    return

                elif S is None:  # already close()'d
                    return

                i = 0
                while True:
                    E = S.pop()
                    if E is None or self._S is None:
                        break
                    yield from self._cb(E)
                    i = (i + 1) % 4
                    if i == 0:
                        yield from asyncio.sleep(0)  # Not sure how necessary.  Ensure we go to the scheduler

                if S.done:
                    _log.debug('Subscription complete %s', self.name)
                    S.close()
                    self._S = None
                    if self._notify_disconnect:
                        E = Finished()
                        yield from self._cb(E)
                    break
        except:
            _log.exception("Error processing Subscription event: %s", LazyRepr(E))
            if self._S is not None:
                self._S.close()
            self._S = None




from __future__ import print_function

import warnings
import os
import sys
import json
import logging
import shutil
from collections import OrderedDict

from . import Server, StaticProvider
from .thread import SharedPV
from .. import nt
from .. import set_debug

_log = logging.getLogger(__name__)

defs = {
    'int': nt.NTScalar('l').wrap(0),
    'uint': nt.NTScalar('L').wrap(0),
    'real': nt.NTScalar('d').wrap(0),
    'str': nt.NTScalar('s').wrap(''),
    'aint': nt.NTScalar('al').wrap([]),
    'auint': nt.NTScalar('aL').wrap([]),
    'areal': nt.NTScalar('ad').wrap([]),
    'astr': nt.NTScalar('as').wrap([]),
    'enum': nt.NTEnum().wrap(0),
}


def getargs():
    from argparse import ArgumentParser

    P = ArgumentParser()
    P.add_argument('-d', '--debug', action='store_const', const=logging.DEBUG, default=logging.INFO)
    P.add_argument('-v', '--verbose', action='store_const', const=logging.DEBUG, default=logging.INFO)
    P.add_argument('-f', '--file', help='Persistence file')

    P.add_argument('pvs', metavar='name=type', nargs='+', help='PV definitions.  types: %s'%(', '.join(defs.keys())))

    return P.parse_args()

def buildMailbox(*args, **kws):
    pv = SharedPV(*args, **kws)

    @pv.put
    def handler(pv, op):
        pv.post(op.value())
        op.done()

    return pv

def main(args):
    db = OrderedDict()
    provider = StaticProvider('soft')

    for pv in args.pvs:
        name, sep, type = pv.partition('=')
        if sep == '':
            print("Invalid definition, missing '=' :", pv)
            sys.exit(1)

        pv = buildMailbox(initial=defs[type])

        provider.add(name, pv)
        db[name] = (type, pv)
        _log.info("Add %s", name)

    if args.file:
        # pre-create to ensure write permission
        OF = open(args.file+'.tmp', 'w')

    if args.file and os.path.exists(args.file):
        with open(args.file, 'r') as F:
            persist = json.load(F)

        if persist['version']!=1:
            warnings.warn('Unknown persist version %s.  Attempting to load'%persist['version'])
        persist = persist['pvs']

        for name, type, iv in persist:
            if name not in db:
                db[name] = (type, buildMailbox(initial=defs[type]))

            _type, pv = db[name]
            pv.post(pv.current().type()(iv))

    try:
        Server.forever(providers=[provider])
    finally:
        persist = []
        for name, (type, pv) in db.items():
            persist.append((name, type, pv.current().todict(None, OrderedDict)))

        if args.file:
            OF.write(json.dumps(OrderedDict([('version',1), ('pvs',persist)]), indent=2))
            OF.flush()
            OF.close()
            shutil.move(args.file+'.tmp', args.file)

if __name__ == '__main__':
    args = getargs()
    set_debug(args.debug)
    logging.basicConfig(level=args.verbose)
    main(args)



import logging
import warnings
import platform
import unittest
import gc
import json
import weakref
import threading

try:
    from Queue import Queue, Full, Empty
except ImportError:
    from queue import Queue, Full, Empty

try:
    from io import StringIO
except ImportError:
    from cString import StringIO

from .utils import RefTestCase, RegularNamedTemporaryFile as NamedTemporaryFile
from ..server import Server, StaticProvider, removeProvider
from ..server.thread import SharedPV, _defaultWorkQueue
from ..client.thread import Context, Disconnected, TimeoutError, RemoteError
from ..nt import NTScalar
from ..gw import App, main, getargs

from .. import _gw

_log = logging.getLogger(__name__)

class TestGC(RefTestCase):
    def test_empty(self):
        class Dummy(object):
            pass
        H = Dummy()
        CLI = _gw.Client(u'pva', {'EPICS_PVA_BROADCAST_PORT': '0',
                                  'EPICS_PVA_SERVER_PORT': '0',
                                  'EPICS_PVAS_INTF_ADDR_LIST': '127.0.0.1',
                                  'EPICS_PVA_ADDR_LIST': '127.0.0.1',
                                  'EPICS_PVA_AUTO_ADDR_LIST': '0'})
        GW = _gw.Provider(u'nulltest', CLI, H)
        removeProvider(u'nulltest')

        self.assertEqual(GW.use_count(), 1)

        h = weakref.ref(H)
        gw = weakref.ref(GW)
        del H
        del GW
        del CLI
        gc.collect()

        self.assertIsNone(h())
        self.assertIsNone(gw())

    def test_server(self):
        class Dummy(object):
            pass
        H = Dummy()
        CLI = _gw.Client(u'pva', {'EPICS_PVA_BROADCAST_PORT': '0',
                                  'EPICS_PVA_SERVER_PORT': '0',
                                  'EPICS_PVAS_INTF_ADDR_LIST': '127.0.0.1',
                                  'EPICS_PVA_ADDR_LIST': '127.0.0.1',
                                  'EPICS_PVA_AUTO_ADDR_LIST': '0'})
        GW = _gw.Provider(u'nulltest', CLI, H)

        try:
            with Server(providers=['nulltest'], isolate=True):
                self.assertFalse(GW.testChannel(b'invalid:pv:name'))
        finally:
            removeProvider(u'nulltest')

        h = weakref.ref(H)
        gw = weakref.ref(GW)
        del H
        del GW
        del CLI
        gc.collect()

        self.assertIsNone(h())
        self.assertIsNone(gw())

class TestLowLevel(RefTestCase):
    timeout = 1

    class Handler(object):
        def testChannel(self, pvname, peer):
            # our permissions check
            if pvname in (b'pv:ro', b'pv:rw'):
                # Add to, and test, channel cache.  Return True if client channel is available
                ret = self.provider.testChannel(b'pv:name')
            else:
                ret = self.provider.BanPV
            _log.debug("GW Search %s from %s -> %s", pvname, peer, ret)
            return ret

        def makeChannel(self, op):
            try:
                # try to create from cache.  Does not add
                chan = op.create(b'pv:name')
                put = False
                if op.name==b'pv:rw':
                    put = True
                    chan.access(put=put, rpc=False, uncached=False)
                _log.debug("GW Create %s put=%s %s for %s of %s", op.name, put, chan, op.account, op.peer)
                return chan
            except:
                _log.exception("Unable to create channel for %s", op.name)

        def audit(self, msg):
            _log.info("AUDIT: %s", msg)

    def setUp(self):
        super(TestLowLevel, self).setUp()

        # upstream server
        self.pv = SharedPV(nt=NTScalar('i'), initial=42)
        self._us_provider = StaticProvider('upstream')
        self._us_provider.add('pv:name', self.pv)

        @self.pv.put
        def put(pv, op):
            _log.debug("PUT %s", op.value())
            pv.post(op.value())
            op.done()

        self._us_server = Server(providers=[self._us_provider], isolate=True)

        # GW client side
        # placed weakref in global registry
        H = self.Handler()
        CLI = _gw.Client(u'pva', self._us_server.conf())
        H.provider = self.gw = _gw.Provider(u'gateway', CLI, H)

        try:
            # GW server side
            self._ds_server = Server(providers=['gateway'], isolate=True)
        finally:
            # don't need this in the global registry anymore.
            # Server holds strong ref.
            removeProvider(u'gateway')

        # downstream client
        self._ds_client = Context('pva', conf=self._ds_server.conf(), useenv=False)

    def tearDown(self):
        self._ds_client.close()
        del self._ds_client

        self._ds_server.stop()
        del self._ds_server

        self._us_server.stop()
        del self._us_provider
        del self._us_server
        del self.pv
        _defaultWorkQueue.sync()
        gc.collect()

        self.assertEqual(self.gw.use_count(), 1)
        gw = weakref.ref(self.gw)
        del self.gw
        gc.collect()

        self.assertIsNone(gw())

        super(TestLowLevel, self).tearDown()

    def test_get(self):
        val = self._ds_client.get('pv:ro', timeout=self.timeout)
        self.assertEqual(val, 42)
        self.pv.post(43)
        # will be delayed by holdoff logic
        val = self._ds_client.get('pv:ro', timeout=self.timeout)
        self.assertEqual(val, 43)

    def test_ban(self):
        with self.assertRaises(TimeoutError):
            self._ds_client.put('invalid', 40, timeout=0.1)
        # TODO: test cache

    def test_put(self):
        with self.assertRaises(RemoteError):
            self._ds_client.put('pv:ro', 40, timeout=self.timeout)

        self.assertEqual(self._ds_client.get('pv:ro', timeout=self.timeout), 42)

        self._ds_client.put('pv:rw', 41, timeout=self.timeout)

        self.assertEqual(self._ds_client.get('pv:ro', timeout=self.timeout), 41)
        self.assertEqual(self._ds_client.get('pv:rw', timeout=self.timeout), 41)

    def test_monitor(self):
        Q1 = Queue(maxsize=4)
        Q2 = Queue(maxsize=4)

        with self._ds_client.monitor('pv:ro', Q1.put, notify_disconnect=True):
            
            self.assertIsInstance(Q1.get(timeout=self.timeout), Disconnected)
            self.assertEqual(42, Q1.get(timeout=self.timeout))

            with self._ds_client.monitor('pv:rw', Q2.put, notify_disconnect=True):
                self.assertIsInstance(Q2.get(timeout=self.timeout), Disconnected)
                self.assertEqual(42, Q2.get(timeout=self.timeout))

                # no activity on first subscription
                with self.assertRaises(Empty):
                    Q2.get(timeout=0.01)

class TestApp(App):
    def __init__(self, args):
        super(TestApp, self).__init__(args)
        self.__evt = threading.Event()

    def abort(self):
        self.__evt.set()

    def sleep(self, dly):
        if self.__evt.wait(dly):
            raise KeyboardInterrupt

class TestHighLevel(RefTestCase):
    timeout = 10

    def setUp(self):
        _log.debug("Enter setUp")
        super(TestHighLevel, self).setUp()

        # upstream server
        self.pv = SharedPV(nt=NTScalar('i'), initial=42)
        self._us_provider = StaticProvider('upstream')
        self._us_provider.add('pv:name', self.pv)

        @self.pv.put
        def put(pv, op):
            _log.debug("PUT %s", op.value())
            pv.post(op.value())
            op.done()

        self._us_server = self._us_conf = None
        self.startServer()
        _log.debug("US server conf: %s", self._us_conf)
        self.assertNotEqual(0, self._us_conf['EPICS_PVA_BROADCAST_PORT'])

        cfile = self._cfile = NamedTemporaryFile('w+')
        json.dump({
            'version':2,
            'clients':[{
                'name':'client1',
                'provider':'pva',
                'addrlist':'127.0.0.1',
                'autoaddrlist':False,
                'bcastport':self._us_conf['EPICS_PVA_BROADCAST_PORT'],
                'serverport':0,
            }],
            'servers':[{
                'name':'server1',
                'clients':['client1'],
                'interface':['127.0.0.1'],
                'addrlist':'127.0.0.1',
                'autoaddrlist':False,
                'bcastport':0,
                'serverport':0,
            }],
        }, cfile)
        cfile.flush()
        with open(cfile.name, 'r') as F:
            _log.debug('GW config')
            _log.debug(F.read())

        # gateway
        args = getargs().parse_args(['--no-ban-local', '-v', cfile.name])

        self._app = TestApp(args)
        self._main = threading.Thread(target=self._app.run, name='GW Main')
        _log.debug("DS server conf: %s", self._app.servers[u'server1_0'].conf())

        # downstream client
        self._ds_client = Context('pva', conf=self._app.servers[u'server1_0'].conf(), useenv=False)

        self._main.start()
        _log.debug("Exit setUp")

    def startServer(self):
        if self._us_server is None:
            if self._us_conf is None:
                _log.debug("Starting server fresh")
                self._us_server = Server(providers=[self._us_provider], isolate=True)
                self._us_conf = self._us_server.conf()
            else:
                _log.debug("Starting server with %s", self._us_conf)
                self._us_server = Server(providers=[self._us_provider], conf=self._us_conf, useenv=False)

    def stopServer(self):
        if self._us_server is not None:
            _log.debug("Stopping server")
            self._us_server.stop()
            self._us_server = None

    def tearDown(self):
        _log.debug("Enter tearDown")
        # downstream client
        self._ds_client.close()
        del self._ds_client

        # gateway
        self._app.abort()
        self._main.join(self.timeout)
        del self._app
        del self._main

        # upstream server
        self.stopServer()
        del self._us_provider
        del self._us_server
        del self.pv
        _defaultWorkQueue.sync()

        super(TestHighLevel, self).tearDown()
        _log.debug("Exit tearDown")

    def test_get(self):
        val = self._ds_client.get('pv:name', timeout=self.timeout)
        self.assertEqual(val, 42)
        self.assertTrue(val.raw.changed('value'))
        # re-read right away to check throttling/holdoff logic
        self.assertEqual(val, 42)

    def test_get_mask(self):
        val = self._ds_client.get('pv:name', timeout=self.timeout, request='timeStamp')
        self.assertEqual(val, 0) # not requested at default
        self.assertFalse(val.raw.changed('value'))

    def test_get_bad_mask(self):
        with self.assertRaisesRegexp(RemoteError, "No field 'nonexistant' Empty field selection"):
            val = self._ds_client.get('pv:name', timeout=self.timeout, request='nonexistant')

    def test_put(self):
        self._ds_client.put('pv:name', 41, timeout=self.timeout)

        val = self._ds_client.get('pv:name', timeout=self.timeout)
        self.assertEqual(val, 41)

    def test_put_bad_mask(self):
        with self.assertRaisesRegexp(RemoteError, "No field 'nonexistant' Empty field selection"):
            self._ds_client.put('pv:name', 41, request='nonexistant', timeout=self.timeout)

    def test_mon(self):
        """Setup a monitor through the GW
        """
        Q = Queue(maxsize=4)
        with self._ds_client.monitor('pv:name', Q.put, notify_disconnect=True):
            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            _log.debug("Wait for initial update")
            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 42)

    def test_mon_disconn(self):
        """See that upstream disconnect propagates
        """
        Q = Queue(maxsize=4)
        with self._ds_client.monitor('pv:name', Q.put, notify_disconnect=True):
            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            with self.assertRaises(Empty):
                Q.get(timeout=0.1)

            _log.debug("Wait for initial update")
            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 42)

            self.stopServer()

            _log.debug("Wait for Disconnected")
            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

    @unittest.skipIf(platform.system()=='Windows',
                     "not responding to searches after restart.  wrong IP?")
    def test_mon_disconn_reconn(self):
        """Start disconnected, the connect and disconnect
        """
        self.stopServer()

        Q = Queue(maxsize=4)
        with self._ds_client.monitor('pv:name', Q.put, notify_disconnect=True):
            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            with self.assertRaises(Empty):
                Q.get(timeout=0.1)

            self.startServer()

            _log.debug("Wait for initial update")
            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 42)

            self.stopServer()

            _log.debug("Wait for Disconnected")
            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            _log.debug("after Disconnected")
            self.pv.post(5)

            with self.assertRaises(Empty):
                Q.get(timeout=0.1)

            self.startServer()

            _log.debug("Wait for reconnect update")
            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 5)

class TestTestServer(RefTestCase):
    conf_template = '''
{
    "version":2,
    "clients":[
        {
            "name":"aclient",
            "provider":"pva",
            "addrlist":"1.2.3.4",
            "autoaddrlist":false,
            "serverport":5085,
            "bcastport":5086
        }
    ],
    "servers":[
        {
            "name":"theserver",
            "clients":["aclient"],
            "interface":["4.3.2.1"],
            "addrlist":"",
            "autoaddrlist":false,
            "serverport":5075,
            "bcastport":5076,
            "getholdoff":1.0,
            "statusprefix":"sts:",
            "access":"%(acf)s",
            "pvlist":"%(pvlist)s",
            "acf_client":"aclient"
        }
    ]
}
'''

    def setUp(self):
        RefTestCase.setUp(self)
        self._files = []
        self._log = StringIO()
        self._handler = logging.StreamHandler(self._log)
        logging.getLogger().addHandler(self._handler)

    def tearDown(self):
        [f.close() for f in self._files]
        logging.getLogger().removeHandler(self._handler)
        RefTestCase.tearDown(self)

    def log(self):
        self._handler.flush()
        self._log.seek(0)
        return self._log.read()

    def write(self, content):
        F = NamedTemporaryFile('w+')
        self._files.append(F)
        F.write(content)
        F.flush()
        return F.name

    def test_ok(self):
        acf = self.write('''
    ASG(DEFAULT) {
        RULE(1, WRITE)
        RULE(1, UNCACHED)
    }
''')
        pvlist = self.write('''
.* ALLOW
''')

        conf = self.write(self.conf_template%{'acf':repr(acf)[1:-1], 'pvlist':repr(pvlist)[1:-1]})

        main(['-T', conf])

    def test_bad_acf(self):
        acf = self.write('''
    ASG(DEFAULT)
        RULE(1, WRITE)
        RULE(1, UNCACHED)
    }
''')
        pvlist = self.write('''
.* ALLOW
''')

        conf = self.write(self.conf_template%{'acf':repr(acf)[1:-1], 'pvlist':repr(pvlist)[1:-1]})

        with self.assertRaises(SystemExit):
            main(['-T', conf])

        self.assertRegex(self.log(), r".*Syntax error.*RULE.*")

    def test_bad_pvlist(self):
        acf = self.write('''
    ASG(DEFAULT) {
        RULE(1, WRITE)
        RULE(1, UNCACHED)
    }
''')
        pvlist = self.write('''
.* ALLW
''')

        conf = self.write(self.conf_template%{'acf':repr(acf)[1:-1], 'pvlist':repr(pvlist)[1:-1]})

        with self.assertRaises(SystemExit):
            main(['-T', conf])

        self.assertRegex(self.log(), r".*Unknown command.*ALLW.*")



# This module defines sub-classes of C extension classes
# which add functionality that is better expressed in python.
# These types are then pushed (by _magic) down into extension
# code where they will be used as the types passed to callbacks.
from ._p4p import TypeBase, ValueBase

__all__ = (
    'Type',
    'Value',
    'Struct',
    'StructArray',
    'Union',
    'UnionArray',
)


def Struct(spec=None, id=None):
    return ('S', id, spec)


def Union(spec=None, id=None):
    return ('U', id, spec)


def StructArray(spec=None, id=None):
    return ('aS', id, spec)


def UnionArray(spec=None, id=None):
    return ('aU', id, spec)


class Type(TypeBase):
    """Type(fields, id=None, base=None)

    :param list fields: A list of tuples describing members of this data structure.
    :param str id: Type label string.
    :param Type base: Copy the fields of `Type`, then amend with the provided fields.

    A definition of a data structure consisting of a list of field names and types,
    as well as an optional type name string (id="").
    Field type specifications are either a string eg. "d" (double precision float),
    or a tuple ("S", None, [fields...) defining a sub-structure. ::

        T = Type([
            ('value', 'I'),
        ])

    Defines a structure with a single field named 'value' with type u32 (unsigned integer width 32-bit).

    An example of defining a sub-structure. ::

        T = Type([
            ('value', ('S', None, [
                ('index', 'i'),
            ])),
        ])

    Type specifier codes:

    ==== =======
    Code Type
    ==== =======
    ?    bool
    s    unicode
    b    s8
    B    u8
    h    s16
    H    u16
    i    i32
    I    u32
    l    i64
    L    u64
    f    f32
    d    f64
    v    variant
    U    union
    S    struct
    ==== =======
    """
    __slots__ = []  # we don't allow custom attributes for now
    __contains__ = TypeBase.has

    def __call__(self, initial=None):
        return Value(self, initial)

    def __iter__(self):
        for k in self.keys():
            yield k

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        return [self[k] for k in self.keys()]

    def __repr__(self):
        S, id, fields = self.aspy()
        return 'Type(%s, id="%s")' % (fields, id)
    __str__ = __repr__

TypeBase._magic(Type)


class Value(ValueBase):
    """Value(type[, initial])

    :param Type type: The `Type` describing the structure to be instanciated
    :param dict initial: A dictionary (or any mapping) which gives initial values for fields.

    Representation of a data structure of a particular :py:class:`Type`.

    Can be created using a :py:class:`Type`, with an optional dict containing initial values. ::

        A = Value(Type([
            ('value', 'I'),
        ]), {
            'value': 42,
        })

    Defines a structure with a single field named 'value' with type u32 (unsigned integer width 32-bit).

    An example of defining a sub-structure. ::

        A = Value(Type([
            ('value', ('S', None, [
                ('index', 'i'),
            ])),
        ]), {
            'value': {'index', 5},
            # 'value.index': 5,  # equivalent
        })

    Defines a structure containing a sub-structure 'value' which has a single field 'index' which is
    a signed 32-bit integer.
    """
    __slots__ = []  # prevent attribute access to invalid fields

    __contains__ = ValueBase.has

    def keys(self):
        return self.type().keys()

    def __iter__(self):
        return iter(self.type())

    def changed(self, *fields):
        """Test if one or more fields have changed.

        A field is considered to have changed if it has been marked as changed,
        or if any of its parent, or child, fields have been marked as changed.
        """
        S = super(Value, self).changed
        for fld in fields or (None,): # no args tests for any change
            if S(fld):
                return True
        return False

    def changedSet(self, expand=False, parents=False):
        """
        :param bool expand: Whether to expand when entire sub-structures are marked as changed.
                            If True, then sub-structures are expanded and only leaf fields will be included.
                            If False, then a direct translation is made, which may include both leaf and sub-structure fields.
        :param bool parents: If True, include fake entries for parent sub-structures with leaf fields marked as changed.
        :returns: A :py:class:`set` of names of those fields marked as changed.

        Return a :py:class:`set` containing the names of all changed fields. ::

            A = Value(Type([
                ('x', 'i'),
                ('z', ('S', None, [
                    ('a', 'i'),
                    ('b', 'i'),
                ])),
            ]), {
            })

            A.mark('z')
            assert A.changedSet(expand=False) == {'z'}         # only shows fields explicitly marked
            assert A.changedSet(expand=True) == {'z.a', 'z.b'} # actually used during network transmission
            A.mark('z.a') # redundant
            assert A.changedSet(expand=False) == {'z', 'z.a'}
            assert A.changedSet(expand=True) == {'z.a', 'z.b'}
            A.unmark('z')
            assert A.changedSet(expand=False) == {'z.a'}
            assert A.changedSet(expand=True) == {'z.a'}
            assert A.changedSet(expand=False, parents=True) == {'z', 'z.a'}
            assert A.changedSet(expand=True, parents=True) == {'z', 'z.a'}


        * expand=False, parents=False gives a direct mapping of the underlying BitSet as it would (get/monitor),
          or have been (put/rpc), moved over the network.
        * expand=True, parents=False gives the effective set of leaf fields which will be moved over the network.
          taking into account the use of whole sub-structure compress/shorthand bits.
        * expand=False, parents=True gives a way of testing if anything changed within a set of interesting fields
          (cf. set.intersect).
        """
        return ValueBase.changedSet(self, expand, parents)

    # TODO: deprecate
    asSet = changedSet

    def clear(self):
        self.mark(None, False)

    __str__ = ValueBase.tostr

    def __repr__(self):
        parts = []

        ID = self.getID()
        if ID!='structure':
            parts.append('id:'+ID)

        try:
            parts.append(repr(self.value))
        except AttributeError: # no .value
            try:
                parts.append(repr(self.get(self.type().keys()[0])))
            except IndexError: # empty Structure
                pass

        return 'Value(%s)'%', '.join(parts)

ValueBase._magic(Value)
